/*
** Module   : TmpSumCleanupContextHook.java
** Abstract : hook for cleaning up tmp-sum records at the end of a session
**
** Copyright (c) 2005-2014, TIMCO Aviation Services, Inc.
** ALL RIGHTS RESERVED.
**
**               TIMCO Aviation Services, Inc.
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 SVL 20090519   @42296 Created initial version. Deletes at the end of a
**                           session two tmp-sum records created upon session
**                           initiation, if they haven't been deleted
**                           normally.
** 002 SVL 20090802   @43474 Retry temp_sum deletion on error. Originally was
**                           designed to handle errors into c3p0 pool.  
** 003 ECF 20131120          Import change.
** 004 SVL 20140210          Use HQLExpression instead of HQL string.
*/

package aero.timco.majic.cfg;

import java.util.*;

import org.apache.commons.logging.*;
import org.hibernate.*;
import org.hibernate.exception.GenericJDBCException;
import org.hibernate.type.*;
import aero.timco.majic.dmo.majic.impl.TempSumImpl;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * This class was intended to solve the following problem with the tmp-sum
 * table in Majic: each user session allocates 2 records in the tmp-sum table
 * for use in various programs while that user's Majic session is active.
 * This is done in menu/init.p. The identifiers of these records are stored in
 * shared variables (cur-sum-id and cur-tmp-id), accessible from every program
 * run after that. When a user logs out normally, these records are deleted in
 * xt/quitsym.p. A problem occurs when the user session is terminated
 * abnormally in P2J, either by a crash, or via user request (Ctrl+C). In
 * either case, the converted quitsym.p program does not run, so we leak
 * tmp-sum records. When we leak too many, it causes login problems. This hook
 * deletes these 2 tmp-sum records if they have not already been deleted by
 * quitsym.p.
 */
public class TmpSumCleanupContextHook
implements InitTermListener
{
   /** HQL used for retrieving tmp-sum record by its identifier */
   private static final String TMP_SUM_RETRIEVE_QUERY =   
         new HQLExpression("from TempSumImpl as tempSum where tempSum.identifier = ", true).
                                                                              toFinalExpression();
   
   /** Majic database name */
   private static final String DATABASE_NAME = "majic";
   
   /** Logger */
   private static final Log LOG =
      LogFactory.getLog(TmpSumCleanupContextHook.class);

   /** Maximum number of temp-sum deletion attempts. */
   private static final int DELETE_ATTEMPTS = 3;

   /** Interval between temp-sum deletion attempts (in milliseconds). */
   private static final int DELETE_INTERVAL = 150;
   
   /** ContextLocal storage of tmp-sum record identifiers */
   private static final ContextLocal<TmpSumContext> context =
   new ContextLocal<TmpSumContext>()
   {
      protected TmpSumContext initialValue()
      {
         return (new TmpSumContext());
      }
   };

   /**
    * This method is invoked when business logic main entry point is about to
    * be invoked. NOP in this implementation.
    */
   public void initialize()
   {
   }
   
   /**
    * This method is invoked when business logic entry point finished its
    * work. Deletes tmp-sum records which identifiers were saved before.
    *
    * @param    t
    *           Error or exception which caused exit from business logic. Can
    *           be <code>null</code>.
    */
   public void terminate(Throwable t)
   {
      TmpSumContext tsc = context.get();
      final integer curSumId = tsc.curSumId;
      final integer curTmpId = tsc.curTmpId;
      
      if (curSumId == null && curTmpId == null)
      {
         if (LOG.isDebugEnabled())
         {
            LOG.debug("Context temp_sum records were cleaned up normally");
         }
         
         return;
      }
      
      Database database = new Database(DATABASE_NAME);
      Persistence persistence = PersistenceFactory.getInstance(database);
      String table =
         DatabaseManager.getBackingTableName(database, TempSumImpl.class);
      
      List<TempSumImpl> records = new ArrayList<TempSumImpl>();
      
      try
      {
         getTempSumsById(persistence, curSumId, records);
         getTempSumsById(persistence, curTmpId, records);

         boolean locked = false;
         
         for (TempSumImpl tempSum : records)
         {
            RecordIdentifier identifier =
               new RecordIdentifier(table, tempSum.getId());
            try
            {
               persistence.lock(LockType.EXCLUSIVE_NO_WAIT, identifier, true);
               locked = true;

               for (int i = 0; i < DELETE_ATTEMPTS; i++)
               {
                  try
                  {
                     persistence.delete(tempSum);
                     break;
                  }
                  catch (GenericJDBCException exc)
                  {
                     // suppress and retry with delay
                     try
                     {
                        Thread.sleep(DELETE_INTERVAL);
                     }
                     catch (InterruptedException e)
                     {
                        // suppress
                     }
                  }

                  if (i == DELETE_ATTEMPTS - 1 && LOG.isWarnEnabled())
                  {
                     LOG.warn("Unable to remove " + identifier);
                  }
               }

               if (LOG.isDebugEnabled())
               {
                  LOG.debug("Removed " + identifier + " at end of context");
               }
            }
            catch (LockUnavailableException exc)
            {
               LOG.error("Unable to lock and remove " + identifier, exc);
            }
            finally
            {
               if (locked)
               {
                  persistence.lock(LockType.NONE, identifier, true);
               }
            }
         }
      }
      catch (PersistenceException e)
      {
         LOG.error("Error cleaning up temp_sum records", e);
      }
   }
   
   /**
    * Stores the value of cur-sum-id variable.
    *
    * @param curSumId
    *        Value to be saved.
    */
   public static void setCurSumId(integer curSumId)
   {
      context.get().curSumId = curSumId;
   }
   
   /**
    * Stores the value of cur-tmp-id variable.
    *
    * @param curTmpId
    *        Value to be saved.
    */
   public static void setCurTmpId(integer curTmpId)
   {
      context.get().curTmpId = curTmpId;
   }
   
   /**
    * Should be invoked after tmp-sum records have been normally cleaned by
    * quitsym.p. Resets identifiers of tmp-sum records to be cleaned what were
    * saved before.    
    */
   public static void normalCleanupDone()
   {
      TmpSumContext tsc = context.get();
      tsc.curSumId = null;
      tsc.curTmpId = null;
   }

   /**
    * Gets tmp-sum record(s) by the given identifier.
    *
    * @param  persistence
    *         Persistence to be used for record retrieval.
    * @param  id
    *         Identifier of tmp-sum record(s) to be retrieved.
    * @param  records
    *         Array that retrieved records should be added to.
    *
    * @throws PersistenceException
    *         If an error has been occured during record retrieval. 
    */
   private void getTempSumsById(Persistence persistence,
                                integer id,
                                List<TempSumImpl> records)
   throws PersistenceException
   {
      if (id != null)
      {
         Object[] vars = new Object[]{id};
         Type[] types = DBUtils.makeTypeArray(vars);
         ScrollableResults rs = persistence.scroll(TMP_SUM_RETRIEVE_QUERY,
                                                   vars,
                                                   types);
         while (rs.next())
         {
            TempSumImpl res = (TempSumImpl) rs.get(0);
            records.add(res);
         }
         rs.close();
      }
   }
   
   /**
    * Container for storing values of cur-sum-id and cur-tmp-id variables.
    */
   private static class TmpSumContext
   {
      /** Stores the value of cur-sum-id variable. */
      private integer curSumId = null;

      /** Stores the value of cur-tmp-id variable. */
      private integer curTmpId = null;
   }
}
